local mod_gui = require("mod-gui")
local Event = require 'utils.event'
local Global = require 'utils.global'
local Gui = require 'utils.gui'
local WPT = require 'maps.mountain_fortress_v3.table'
local low_time_left_label_color = {r = 1}
local completed_label_color = {g = 1}

local Public = {}
local this = {
	players = {},
	Linked = {},
	required = {},
	accumulated = {},
	level_started_at = 0,
    time_modifier = 1.4,
	points_per_second_start = 5,
	points_per_second_level_subtract = 0.2,
	coins = 0,
	level = 1,
	game_end = false,
	levels =
	{
	  -- 1
	  {
		requirements =
		{
		  {name = "stone", count = 30},
		},
		time = 240
	  },

	  -- 2
	  {
		requirements =
		{
		  {name = "iron-plate", count = 30}
		},
		time = 300
	  },

	  -- 3
	  {
		requirements =
		{
		  {name = "iron-plate", count = 30},
		  {name = "copper-plate", count = 30}
		},
		time = 300
	  },

	  -- 4
	  {
		requirements =
		{
		  {name = "iron-plate", count = 30},
		  {name = "iron-gear-wheel", count = 30}
		},
		time = 300
	  },

	  -- 5
	  {
		requirements =
		{
		  {name = "iron-plate", count = 40},
		  {name = "iron-gear-wheel", count = 30},
		  {name = "copper-cable", count = 40}
		},
		time = 300
	  },

	  -- 6
	  {
		requirements =
		{
		  {name = "iron-plate", count = 40},
		  {name = "iron-gear-wheel", count = 30},
		  {name = "electronic-circuit", count = 30}
		},
		time = 300
	  },

	  -- 7
	  {
		requirements =
		{
		  {name = "iron-plate", count = 40},
		  {name = "iron-gear-wheel", count = 30},
		  {name = "electronic-circuit", count = 30},
		  {name = "automation-science-pack", count = 10},
		  {name = "firearm-magazine", count = 50}
		},
		time = 300
	  },

	  -- 8
	  {
		requirements =
		{
		  {name = "iron-plate", count = 40},
		  {name = "iron-gear-wheel", count = 30},
		  {name = "electronic-circuit", count = 30},
		  {name = "automation-science-pack", count = 10},
		  {name = "transport-belt", count = 20}
		},
		time = 300
	  },

	  -- 9
	  {
		requirements =
		{
		  {name = "iron-plate", count = 50},
		  {name = "iron-gear-wheel", count = 30},
		  {name = "electronic-circuit", count = 30},
		  {name = "automation-science-pack", count = 10},
		  {name = "transport-belt", count = 20},
		  {name = "inserter", count = 20}
		},
		time = 300
	  },

	  -- 10
	  {
		requirements =
		{
		  {name = "iron-plate", count = 50},
		  {name = "iron-gear-wheel", count = 30},
		  {name = "electronic-circuit", count = 30},
		  {name = "automation-science-pack", count = 10},
		  {name = "logistic-science-pack", count = 10}
		},
		time = 300
	  },

	  -- 11
	  {
		requirements =
		{
		  {name = "iron-plate", count = 50},
		  {name = "iron-gear-wheel", count = 30},
		  {name = "electronic-circuit", count = 30},
		  {name = "automation-science-pack", count = 20},
		  {name = "logistic-science-pack", count = 20},
		  {name = "fast-transport-belt", count = 10}
		},
		time = 300
	  },

	  -- 12
	  {
		requirements =
		{
		  {name = "iron-plate", count = 50},
		  {name = "iron-gear-wheel", count = 30},
		  {name = "electronic-circuit", count = 30},
		  {name = "automation-science-pack", count = 20},
		  {name = "logistic-science-pack", count = 20},
		  {name = "steel-plate", count = 10}
		},
		time = 420
	  },

	  -- 13
	  {
		requirements =
		{
		  {name = "iron-gear-wheel", count = 30},
		  {name = "electronic-circuit", count = 30},
		  {name = "automation-science-pack", count = 20},
		  {name = "logistic-science-pack", count = 20},
		  {name = "steel-plate", count = 15},
		  {name = "piercing-rounds-magazine", count = 50}
		},
		time = 300
	  },

	  -- 14
	  {
		requirements =
		{
		  {name = "iron-gear-wheel", count = 30},
		  {name = "electronic-circuit", count = 30},
		  {name = "automation-science-pack", count = 20},
		  {name = "logistic-science-pack", count = 20},
		  {name = "steel-plate", count = 15},
		  {name = "grenade", count = 15},
		  {name = "plastic-bar", count = 5}
		},
		time = 600
	  },

	  -- 15
	  {
		requirements =
		{
		  {name = "electronic-circuit", count = 30},
		  {name = "automation-science-pack", count = 20},
		  {name = "logistic-science-pack", count = 20},
		  {name = "steel-plate", count = 15},
		  {name = "advanced-circuit", count = 10}
		},
		time = 500
	  },

	  -- 16
	  {
		requirements =
		{
		  {name = "automation-science-pack", count = 20},
		  {name = "logistic-science-pack", count = 20},
		  {name = "steel-plate", count = 15},
		  {name = "advanced-circuit", count = 10},
		  {name = "solid-fuel", count = 10}
		},
		time = 500
	  },

	  -- 17
	  {
		requirements =
		{
		  {name = "automation-science-pack", count = 20},
		  {name = "logistic-science-pack", count = 20},
		  {name = "advanced-circuit", count = 10},
		  {name = "engine-unit", count = 10},
		  {name = "solid-fuel", count = 10}
		},
		time = 500
	  },

	  -- 18
	  {
		requirements =
		{
		  {name = "automation-science-pack", count = 20},
		  {name = "logistic-science-pack", count = 20},
		  {name = "chemical-science-pack", count = 20},
		  {name = "grenade", count = 20}
		},
		time = 500
	  },

	  -- 19
	  {
		requirements =
		{
		  {name = "automation-science-pack", count = 30},
		  {name = "logistic-science-pack", count = 30},
		  {name = "chemical-science-pack", count = 30},
		  {name = "stone-wall", count = 20}
		},
		time = 500
	  },

	  -- 20
	  {
		requirements =
		{
		  {name = "automation-science-pack", count = 40},
		  {name = "logistic-science-pack", count = 40},
		  {name = "chemical-science-pack", count = 40},
		  {name = "stone-wall", count = 20},
		  {name = "grenade", count = 20}
		},
		time = 600
	  },

	  -- 21
	  {
		requirements =
		{
		  {name = "automation-science-pack", count = 50},
		  {name = "logistic-science-pack", count = 50},
		  {name = "chemical-science-pack", count = 50},
		  {name = "military-science-pack", count = 50}
		},
		time = 700
	  }
	},
    activate_custom_buttons = true,--激活自定义按钮
    bottom_right = true,
    bottom_quickbar_button_112 = {},
	players_donationitems_main_location = {},
	
	
}
Global.register(
    this,
    function(t)
        this = t
    end
)

function Public.reset()

	this.required = {}
	this.players = {}
	this.accumulated = {}
	this.level_started_at = game.tick
    this.time_modifier = 1.4
	this.points_per_second_start = 5
	this.points_per_second_level_subtract = 0.2
	this.level = 1
	this.coins = 0
	this.game_end = false
	this.levels =
	{
	  -- 1
	  {
		requirements =
		{
		  {name = "stone", count = 30},
		},
		time = 240
	  },

	  -- 2
	  {
		requirements =
		{
		  {name = "iron-plate", count = 30}
		},
		time = 300
	  },

	  -- 3
	  {
		requirements =
		{
		  {name = "iron-plate", count = 30},
		  {name = "copper-plate", count = 30}
		},
		time = 300
	  },

	  -- 4
	  {
		requirements =
		{
		  {name = "iron-plate", count = 30},
		  {name = "iron-gear-wheel", count = 30}
		},
		time = 300
	  },

	  -- 5
	  {
		requirements =
		{
		  {name = "iron-plate", count = 40},
		  {name = "iron-gear-wheel", count = 30},
		  {name = "copper-cable", count = 40}
		},
		time = 300
	  },

	  -- 6
	  {
		requirements =
		{
		  {name = "iron-plate", count = 40},
		  {name = "iron-gear-wheel", count = 30},
		  {name = "electronic-circuit", count = 30}
		},
		time = 300
	  },

	  -- 7
	  {
		requirements =
		{
		  {name = "iron-plate", count = 40},
		  {name = "iron-gear-wheel", count = 30},
		  {name = "electronic-circuit", count = 30},
		  {name = "automation-science-pack", count = 10},
		  {name = "firearm-magazine", count = 50}
		},
		time = 300
	  },

	  -- 8
	  {
		requirements =
		{
		  {name = "iron-plate", count = 40},
		  {name = "iron-gear-wheel", count = 30},
		  {name = "electronic-circuit", count = 30},
		  {name = "automation-science-pack", count = 10},
		  {name = "transport-belt", count = 20}
		},
		time = 300
	  },

	  -- 9
	  {
		requirements =
		{
		  {name = "iron-plate", count = 50},
		  {name = "iron-gear-wheel", count = 30},
		  {name = "electronic-circuit", count = 30},
		  {name = "automation-science-pack", count = 10},
		  {name = "transport-belt", count = 20},
		  {name = "inserter", count = 20}
		},
		time = 300
	  },

	  -- 10
	  {
		requirements =
		{
		  {name = "iron-plate", count = 50},
		  {name = "iron-gear-wheel", count = 30},
		  {name = "electronic-circuit", count = 30},
		  {name = "automation-science-pack", count = 10},
		  {name = "logistic-science-pack", count = 10}
		},
		time = 300
	  },

	  -- 11
	  {
		requirements =
		{
		  {name = "iron-plate", count = 50},
		  {name = "iron-gear-wheel", count = 30},
		  {name = "electronic-circuit", count = 30},
		  {name = "automation-science-pack", count = 20},
		  {name = "logistic-science-pack", count = 20},
		  {name = "fast-transport-belt", count = 10}
		},
		time = 300
	  },

	  -- 12
	  {
		requirements =
		{
		  {name = "iron-plate", count = 50},
		  {name = "iron-gear-wheel", count = 30},
		  {name = "electronic-circuit", count = 30},
		  {name = "automation-science-pack", count = 20},
		  {name = "logistic-science-pack", count = 20},
		  {name = "steel-plate", count = 10}
		},
		time = 420
	  },

	  -- 13
	  {
		requirements =
		{
		  {name = "iron-gear-wheel", count = 30},
		  {name = "electronic-circuit", count = 30},
		  {name = "automation-science-pack", count = 20},
		  {name = "logistic-science-pack", count = 20},
		  {name = "steel-plate", count = 15},
		  {name = "piercing-rounds-magazine", count = 50}
		},
		time = 300
	  },

	  -- 14
	  {
		requirements =
		{
		  {name = "iron-gear-wheel", count = 30},
		  {name = "electronic-circuit", count = 30},
		  {name = "automation-science-pack", count = 20},
		  {name = "logistic-science-pack", count = 20},
		  {name = "steel-plate", count = 15},
		  {name = "grenade", count = 15},
		  {name = "plastic-bar", count = 5}
		},
		time = 600
	  },

	  -- 15
	  {
		requirements =
		{
		  {name = "electronic-circuit", count = 30},
		  {name = "automation-science-pack", count = 20},
		  {name = "logistic-science-pack", count = 20},
		  {name = "steel-plate", count = 15},
		  {name = "advanced-circuit", count = 10}
		},
		time = 500
	  },

	  -- 16
	  {
		requirements =
		{
		  {name = "automation-science-pack", count = 20},
		  {name = "logistic-science-pack", count = 20},
		  {name = "steel-plate", count = 15},
		  {name = "advanced-circuit", count = 10},
		  {name = "solid-fuel", count = 10}
		},
		time = 500
	  },

	  -- 17
	  {
		requirements =
		{
		  {name = "automation-science-pack", count = 20},
		  {name = "logistic-science-pack", count = 20},
		  {name = "advanced-circuit", count = 10},
		  {name = "engine-unit", count = 10},
		  {name = "solid-fuel", count = 10}
		},
		time = 500
	  },

	  -- 18
	  {
		requirements =
		{
		  {name = "automation-science-pack", count = 20},
		  {name = "logistic-science-pack", count = 20},
		  {name = "chemical-science-pack", count = 20},
		  {name = "grenade", count = 20}
		},
		time = 500
	  },

	  -- 19
	  {
		requirements =
		{
		  {name = "automation-science-pack", count = 30},
		  {name = "logistic-science-pack", count = 30},
		  {name = "chemical-science-pack", count = 30},
		  {name = "stone-wall", count = 20}
		},
		time = 500
	  },

	  -- 20
	  {
		requirements =
		{
		  {name = "automation-science-pack", count = 40},
		  {name = "logistic-science-pack", count = 40},
		  {name = "chemical-science-pack", count = 40},
		  {name = "stone-wall", count = 20},
		  {name = "grenade", count = 20}
		},
		time = 600
	  },

	  -- 21
	  {
		requirements =
		{
		  {name = "automation-science-pack", count = 50},
		  {name = "logistic-science-pack", count = 50},
		  {name = "chemical-science-pack", count = 50},
		  {name = "military-science-pack", count = 50}
		},
		time = 700
	  }
	}
    this.activate_custom_buttons = true--激活自定义按钮
    this.bottom_right = true
    this.bottom_quickbar_button_112 = {}
	this.players_donationitems_main_location = {}
	
	game.forces.player.technologies["logistic-robotics"].researched = true
	local level = this.levels[1]
	for index, item in pairs(level.requirements) do
	  this.accumulated[item.name] = 0
	  this.required[item.name] = item.count
	end
	
end

Event.on_init(
    function()
		this.required = {}
		this.accumulated = {}
		this.labels = {}
		this.coins = 0
		this.players = {}
		this.level_started_at = game.tick
		game.forces.player.technologies["logistic-robotics"].researched = true
		local level = this.levels[1]
		for index, item in pairs(level.requirements) do
		  this.accumulated[item.name] = 0
		  this.required[item.name] = item.count
		end
    end
)
local accumulate_items = function()--累积物品
  local accumulated = this.accumulated
	for index, player in pairs(game.connected_players) do
		local character = player.character
		if character and character.valid then
			local player_trash = player.get_inventory(defines.inventory.character_trash)
			if player_trash and player_trash.valid then
				if not player_trash.is_empty() then
					local contents = player_trash.get_contents()
					for item_name, count in pairs(contents) do
						if accumulated[item_name] then
							local count_to_consume = math.min(this.required[item_name] - accumulated[item_name], count)
							if count_to_consume ~= 0 then
								player_trash.remove{name = item_name, count = count_to_consume}
								accumulated[item_name] = accumulated[item_name] + count_to_consume
							end
						end
					end
				end
			end
		end
	end
end

function formattime(ticks)
  local seconds = ticks / 60
  local hours   = math.floor((seconds)/60/60)
  local minutes = math.floor((seconds)/60 - 60*hours)
  local seconds = math.floor(seconds - 60*minutes - 60*60*hours)
  return string.format("时间：%d时%02d分%02d秒", hours,minutes, seconds)
end

function get_time_left()
  --return this.level_started_at + this.time_modifier * this.level * 60 - game.tick
  return this.level_started_at + this.time_modifier * this.levels[1].time * 60 - game.tick
end

local function update_supply_main_player(frame, player)
	local table1 = frame.inner.table1
	local time_now_label = table1["time-now"]
	local time_left_label = table1["time-left"]

	local required_items_flow = frame['required-items-flow']
	local required_items_table = table1['required-items-flow']['required-items-table']
	local next_level_table = table1['next-level-flow']['next-level-table']
	
	
	required_items_table.clear()
	next_level_table.clear()
	
	
	if time_now_label then
	--game.print('更新现在时间' .. formattime(game.tick))
		time_now_label.caption = formattime(game.tick) 
		time_left_label.caption = formattime(get_time_left()) 


		local level = this.levels[1]
		for index, item in pairs(level.requirements) do
			local accumulated = this.accumulated[item.name]
			local sprite = required_items_table.add{type = "sprite", sprite = "item/"..item.name, style = "small_text_image"}
			required_items_table.add{type = "label", caption = {"", game.item_prototypes[item.name].localised_name, {"colon"}}}
			local label = required_items_table.add{type = "label", caption = accumulated .. "/" .. item.count}
			if accumulated == item.count then
				label.style.font_color = completed_label_color
			end
		end
			
		local next_level = this.levels[1 + 1]
		if next_level then
			for index, item in pairs(next_level.requirements) do
				local sprite = next_level_table.add{type = "sprite", sprite = "item/"..item.name, style = "small_text_image"}
				next_level_table.add{type = "label", caption = {"", game.item_prototypes[item.name].localised_name, {"colon"}}}
				next_level_table.add{type = "label", caption = item.count}
			end
		end
		
		table1['add-coins'].caption = {"add-coin", this.coins}
		table1['level'].caption = {"levels", this.level}
	end
	
	local time_left = get_time_left()
	if time_left < 60 * 30 then
		time_left_label.style.font_color = low_time_left_label_color--time_left_label
		local text1 = '物质任务剩余：' .. math.floor(time_left/60) .. '秒！！'
		for index, player in pairs(game.connected_players) do
			player.surface.create_entity({name = 'flying-text', position = player.position, text = text1, color = {255, 0, 0}})
			
			
		end
	else
		time_left_label.style.font_color = {r = 1,g = 1,b = 1}
	end
	
	
end



local function draw_supply_main_player(player)
	if not player or not player.valid then
		return
	end
	local panel = player.gui.relative
	local frame = panel["supplymain"]
	local time_left = get_time_left()
	if not frame then
		local anchor = {gui = defines.relative_gui_type.controller_gui, position = defines.relative_gui_position.right}
		frame = panel.add {
			type = 'frame',
			name = "supplymain",
			style = mod_gui.frame_style,
			direction = "vertical",
			caption ='【供应任务界面】',
			anchor = anchor,
		}
		local inner = frame.add{type = "frame",name = 'inner', direction = "vertical", style = "inside_shallow_frame"}
		
		local table1 = inner.add{type = "table",name = 'table1', column_count = 1, style = "bordered_table"}
		table1.style.margin = 4
		local time_now_label = table1.add{type = "label", name = "time-now", caption = formattime(game.tick)}
		local time_left_label = table1.add{type = "label", name = "time-left", caption = {"time-left", util.formattime(time_left)}}
		table1.add{type = "label", name = 'add-coins', caption = {"add-coin", this.coins}}
		table1.add{type = "label", name = 'level', caption = {"level", this.level}}
		
		local required_items_flow = table1.add{type = "flow", name = 'required-items-flow', direction = "vertical"}

		required_items_flow.add{type = "label", name = 'required-items', caption = {"required-items"}, style = "caption_label"}

		local item_prototypes = game.item_prototypes
		local accumulated = this.accumulated

		local required_items_table = required_items_flow.add{type = "table",name = 'required-items-table', column_count = 3}
		required_items_table.style.column_alignments[3] = "right"
		local level = this.levels[1]
		for index, item in pairs(level.requirements) do
			local accumulated = accumulated[item.name]
			local sprite = required_items_table.add{type = "sprite", sprite = "item/"..item.name, style = "small_text_image"}
			required_items_table.add{type = "label", caption = {"", item_prototypes[item.name].localised_name, {"colon"}}}
			local label = required_items_table.add{type = "label", caption = accumulated .. "/" .. item.count}
			if accumulated == item.count then
				label.style.font_color = completed_label_color
			end
		end

		local next_level = this.levels[1 + 1]
		if next_level then
			local next_level_flow = table1.add{type = "flow",name = 'next-level-flow', direction = "vertical"}
			next_level_flow.add{type= "label", name = 'next-level-label',caption = {"next-level"}, style = "caption_label"}
			local next_level_table = next_level_flow.add{type = "table",name = 'next-level-table', column_count = 3}
			next_level_table.style.column_alignments[3] = "right"
			for index, item in pairs(next_level.requirements) do
				local sprite = next_level_table.add{type = "sprite", sprite = "item/"..item.name, style = "small_text_image"}
				next_level_table.add{type = "label", caption = {"", item_prototypes[item.name].localised_name, {"colon"}}}
				next_level_table.add{type = "label", caption = item.count}
			end
		end
		
		return
	end
	update_supply_main_player(frame, player)

	
	


end


--tick
local function tick()
	local ticker = game.tick
	
	if ticker % 60 ~= 0 then return end
	if this.game_end then return end

	
	accumulate_items()
	for k, player in pairs (game.players) do
	  draw_supply_main_player(player)
	end
	

		
	local result = true
	local level = this.levels[1]
	for index, item in pairs(level.requirements) do
		local accumulated = this.accumulated[item.name]
		if accumulated < item.count then
			result = false
		end
	end
	-- if result then
		-- for k, player in pairs (game.players) do
			-- get_next_level_button(player).enabled = true
		-- end
	-- end
	local nexttime =  math.floor(get_time_left()/60)
	
	if nexttime < 0 then nexttime = 0 end
	if result == false and nexttime == 0  then
		for k, player in pairs (game.players) do
			
			local supplymain = player.gui.relative["supplymain"]
			if supplymain and supplymain.valid then supplymain.destroy() end
			--player.gui.top.clear()
			--player.gui.left.clear()
		end
		
		
		for index, player in pairs(game.connected_players) do
			local character = player.character
			if character and character.valid then
				character.damage(9999999999, 'enemy', 'explosion')
			end
		end
		game.print('物质任务挑战失败！！！全体阵亡.......')
		game.print('本次任务已重启.....')
		this.level_started_at = game.tick
	elseif result and nexttime == 0  then
		--return true
		this.required = {}
		this.accumulated = {}
		this.level = this.level + 1
		table.remove(this.levels,1)
		this.level_started_at = game.tick
		for index, item in pairs(this.levels[1].requirements) do
		  this.accumulated[item.name] = 0
		  this.required[item.name] = item.count
		end
		
		givecoin = math.random(100, 200) * this.level
		
		for index, player in pairs(game.connected_players) do
			local character = player.character
			if character and character.valid then
				character.insert{name = "coin", count = givecoin}
			end
			game.print('物资任务挑战成功->所有玩家奖励金币:' .. givecoin)
		end
		this.coins = this.coins + givecoin
		table.insert(this.levels,
		  {
			requirements =
			{
			  {name = "automation-science-pack", count =  70 + this.level},
			  {name = "logistic-science-pack", count =  60 + this.level},
			  {name = "military-science-pack", count =  50 + this.level},
			  {name = "chemical-science-pack", count =  40 + this.level},
			  {name = "production-science-pack", count =  30 + this.level},
			  {name = "utility-science-pack", count =  20 + this.level},
			  {name = "space-science-pack", count =  10 + this.level},
			},
			time = 60 + this.level/2
		  }
		)
		  
	end	  
end

--on_gui_confirmed
Event.add(
    defines.events.on_gui_confirmed,
	function(event)
		-- element :: LuaGuiElement: The confirmed element.
		-- player_index :: uint: The player who did the confirming.
		-- alt :: boolean: If alt was pressed.
		-- control :: boolean: If control was pressed.
		-- shift :: boolean: If shift was pressed.
		local gui_name = 'LinkedPassword-'
		local player = game.players[event.player_index]
		if string.sub(event.element.name, 0, #gui_name) == gui_name then
			local unit_number = tonumber(string.sub(event.element.name, #gui_name + 1, #event.element.name))
			local Password = tonumber(event.element.text)
			if Password > 4294967295 then
				player.print('关联箱密码范围 0 ~ 4294967295')
				return
			end
			
			if this.players[player.name].Linked and this.players[player.name].Linked.valid then 
				if  unit_number == this.players[player.name].Linked.unit_number then
					this.players[player.name].Linked.link_id = Password
					player.print({'set_Password',Password},{r = 0.75, g = 0.0, b = 0.0})
				end
			end
		end
	end
)
--on_gui_opened
Event.add(
    defines.events.on_gui_opened,
	function(event)
		-- player_index :: uint: The player.
		-- gui_type :: defines.gui_type: The GUI type that was opened.
		-- entity :: LuaEntity (optional): The entity that was opened
		-- item :: LuaItemStack (optional): The item that was opened
		-- equipment :: LuaEquipment (optional): The equipment that was opened
		-- other_player :: LuaPlayer (optional): The other player that was opened
		-- element :: LuaGuiElement (optional): The custom GUI element that was opened
		local player = game.players[event.player_index]
		if event.gui_type == defines.gui_type.entity then
			local entity = event.entity
			if entity.name == 'linked-chest' then
				local gui_name = 'LinkedPassword-'
				local anchor = {gui = defines.relative_gui_type.linked_container_gui, position = defines.relative_gui_position.top}
				local panel = player.gui.relative
				local frame = panel["set-linked-chest-password"]
				if not frame then
					frame = panel.add {
						type = 'frame',
						name = "set-linked-chest-password",
						style = mod_gui.frame_style,
						direction = "vertical",
						caption = '设置关联箱密码',
						anchor = anchor,
					}
					
					if this.players[player.name] == nil then
						this.players[player.name] = {}
						this.players[player.name].Linked = entity 
					else
						this.players[player.name].Linked = entity 
					end
					
					
					local LinkedPassword = frame.add{type = "textfield",name = gui_name .. entity.unit_number, text  = this.players[player.name].Linked.link_id, tooltip = "Enter确认密码\n关联箱密码范围 0 ~ 4294967295"}
					
					LinkedPassword.style.height = 40
					--LinkedPassword.style.width = 400
					LinkedPassword.style.top_margin = 4
					LinkedPassword.style.bottom_margin = 4
					LinkedPassword.numeric = true
					LinkedPassword.allow_decimal = true
					LinkedPassword.allow_negative  = false
					
				else
					if this.players[player.name] == nil then
						this.players[player.name] = {}
						this.players[player.name].Linked = entity 
					else
						this.players[player.name].Linked = entity 
					end
					frame.clear()
					local LinkedPassword = frame.add{type = "textfield",name = gui_name .. entity.unit_number, text  = this.players[player.name].Linked.link_id, tooltip = "Enter确认密码\n关联箱密码范围 0 ~ 4294967295"}
					
					LinkedPassword.style.height = 40
					--LinkedPassword.style.width = 400
					LinkedPassword.style.top_margin = 4
					LinkedPassword.style.bottom_margin = 4
					LinkedPassword.numeric = true
					LinkedPassword.allow_decimal = true
					LinkedPassword.allow_negative  = false
				end
			end
		end
	end
)
--on_gui_closed
-- Event.add(
    -- defines.events.on_gui_closed,
	-- function(event)
		-- -- player_index :: uint: The player.
		-- -- gui_type :: defines.gui_type: The GUI type that was open.
		-- -- entity :: LuaEntity (optional): The entity that was open
		-- -- item :: LuaItemStack (optional): The item that was open
		-- -- equipment :: LuaEquipment (optional): The equipment that was open
		-- -- other_player :: LuaPlayer (optional): The other player that was open
		-- -- element :: LuaGuiElement (optional): The custom GUI element that was open
		-- -- technology :: LuaTechnology (optional): The technology that was automatically selected when opening the research GUI
		-- -- tile_position :: TilePosition (optional): The tile position that was open
	


	-- end
-- )



--on_gui_click
-- Event.add(
    -- defines.events.on_gui_click,
    -- function(event)
		-- if not event then
			-- return
		-- end
		-- game.print('直接界面' .. event.element.name)
		-- game.print('父级' .. event.element.parent.name)
    -- end
-- )
Event.on_nth_tick(5, tick)

return Public